// Licensed to Apache Software Foundation (ASF) under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Apache Software Foundation (ASF) licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

package flow

// Transmit provides a helper function to connect the current component with the downstream component.
// It should be run in another goroutine.
func Transmit(state *ComponentState, downstream Inlet, current Outlet) {
	defer state.Done()
	for elem := range current.Out() {
		downstream.In() <- elem
	}
	close(downstream.In())
}

// TryExactTimestamp tries to get the timestamp from the item.
func TryExactTimestamp(item any) StreamRecord {
	if r, ok := item.(StreamRecord); ok {
		return r
	}
	type timestampExtractor interface {
		TimestampMillis() int64
	}
	// otherwise, check if we can extract timestamp
	if extractor, ok := item.(timestampExtractor); ok {
		return NewStreamRecord(item, extractor.TimestampMillis())
	}
	return NewStreamRecordWithoutTS(item)
}
