// Licensed to Apache Software Foundation (ASF) under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Apache Software Foundation (ASF) licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

syntax = "proto3";

package banyandb.database.v1;

import "banyandb/common/v1/common.proto";
import "google/protobuf/timestamp.proto";

option go_package = "github.com/apache/skywalking-banyandb/api/proto/banyandb/database/v1";
option java_package = "org.apache.skywalking.banyandb.database.v1";

enum Role {
  ROLE_UNSPECIFIED = 0;
  ROLE_META = 1;
  ROLE_DATA = 2;
  ROLE_LIAISON = 3;
}

message Node {
  common.v1.Metadata metadata = 1;
  repeated Role roles = 2;
  string grpc_address = 3;
  string http_address = 4;
  google.protobuf.Timestamp created_at = 5;
  // labels is a set of key-value pairs to describe the node.
  map<string, string> labels = 6;
  string property_repair_gossip_grpc_address = 7;
  string property_schema_grpc_address = 8;
  string property_schema_gossip_grpc_address = 9;
}

message Shard {
  uint64 id = 1;
  common.v1.Metadata metadata = 2;
  common.v1.Catalog catalog = 3;
  string node = 4;
  uint32 total = 5;
  google.protobuf.Timestamp updated_at = 6;
  google.protobuf.Timestamp created_at = 7;
}
