// Licensed to Apache Software Foundation (ASF) under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Apache Software Foundation (ASF) licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

// Package checksum provides functions for computing checksums algorithms and verifying.
package checksum

import "io"

// Verifier defines the interface for computing and verifying checksums.
type Verifier interface {
	// Wrap returns an io.ReadCloser that transparently verifies the checksum
	// when the returned reader is closed. This enables streaming verification
	// without buffering the entire content in memory.
	Wrap(io.ReadCloser, string) io.ReadCloser

	// ComputeAndWrap returns a reader that computes the checksum while reading.
	ComputeAndWrap(r io.Reader) (wrappedReader io.Reader, getHash func() (string, error))
}
