// Licensed to Apache Software Foundation (ASF) under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Apache Software Foundation (ASF) licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

syntax = "proto3";

package banyandb.common.v1;

import "google/api/annotations.proto";

option go_package = "github.com/apache/skywalking-banyandb/api/proto/banyandb/common/v1";
option java_package = "org.apache.skywalking.banyandb.common.v1";

// APIVersion is the version of the API
message APIVersion {
  // version is the version of the API
  string version = 1;
  // revision is the commit hash of the API
  string revision = 2;
}

// GetAPIVersionRequest is the request for GetAPIVersion
message GetAPIVersionRequest {
  // empty
}

// GetAPIVersionResponse is the response for GetAPIVersion
message GetAPIVersionResponse {
  // version is the version of the API
  APIVersion version = 1;
}

// Service is the service for the API
service Service {
  // GetAPIVersion returns the version of the API
  rpc GetAPIVersion(GetAPIVersionRequest) returns (GetAPIVersionResponse) {
    option (google.api.http) = {get: "/v1/common/api/version"};
  }
}
