/*
 * Licensed to Apache Software Foundation (ASF) under one or more contributor
 * license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright
 * ownership. Apache Software Foundation (ASF) licenses this file to you under
 * the Apache License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

html {
  /* ==================
            color
    ==================== */
  --color-main: #6e38f7; /* main color */
  --el-color-primary-light-3: #936bf8 !important;
  --el-color-primary: var(--color-main) !important;
  --el-border-color-hover: var(--color-main) !important;
  --el-button-hover-bg-color: var(--color-main-background) !important;
  --el-button-hover-border-color: var(--color-main) !important;
  --color-main-background: rgb(241, 235, 254);
  --color-background: #e7ebfc;

  --color-top-background: #1d1f2a;
  --color-nav-background: #4b4c59;

  --color-main-font: #303133;
  --color-general-font: #606266;
  --color-secondary-font: #909399;
  --color-placeholder-font: #c0c4cc;

  --color-first-border: #dcdfe6;
  --color-secondary-border: #e4e7ed;
  --color-Tertiary-border: #ebeef5;
  --color-Quad-border: #f2f6fc;

  --color-black: #000000;
  --color-white: #ffffff;

  --color-select: #ecf5ff;

  /* ==================
            font
    ==================== */
  --weight-mini: 500; /* font-weight */
  --weight-lt: 600;
  --size-mini: 0.8em; /* font-size */
  --size-lt: 1.1em;
  --size-title: 1.2em;
  --size-big: 1.4em;
  --font-family-main: 'Lucida Sans', 'Lucida Sans Regular', 'Lucida Sans Unicode', Verdana, sans-serif;
  --font-color: #3d444f;
  --text-color: #fff;
  --theme-background: #fff;
  --active-color: var(--el-color-primary);
  --disabled-color: #ccc;
  --scrollbar-track: #e3dede;
  --scrollbar-thumb: #aaa5a5;
}

div {
  box-sizing: border-box;
}
