// Licensed to Apache Software Foundation (ASF) under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Apache Software Foundation (ASF) licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

package convert

import (
	"unsafe"
)

var emptyStringBytes = []byte{}

// StringToBytes converts string to bytes.
// It works well until the referenced memory won’t be changed.
func StringToBytes(s string) (b []byte) {
	if s == "" {
		return emptyStringBytes
	}
	return unsafe.Slice(unsafe.StringData(s), len(s))
}

// BytesToString converts bytes to string.
// It works well until the referenced memory won’t be changed.
func BytesToString(b []byte) string {
	if len(b) == 0 {
		return ""
	}
	return unsafe.String(unsafe.SliceData(b), len(b))
}
