// Licensed to Apache Software Foundation (ASF) under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Apache Software Foundation (ASF) licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//	http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

// Package flags implements injected flags to run tests.
package flags

import (
	"strconv"
	"time"
)

var (
	eventuallyTimeout string

	neverTimeout string

	consistentlyTimeout string

	requireKTM string

	// EventuallyTimeout is the timeout of async time cases execution.
	EventuallyTimeout time.Duration

	// ConsistentlyTimeout is the timeout of async time cases execution.
	ConsistentlyTimeout time.Duration

	// NeverTimeout is the timeout of async time cases execution.
	NeverTimeout time.Duration

	// LogLevel is the log level of test cases.
	LogLevel = "debug"

	// RequireKTM indicates whether KTM integration tests must run and should fail instead of skipping when requirements are missing.
	RequireKTM bool
)

func init() {
	if eventuallyTimeout == "" {
		eventuallyTimeout = "10s"
	}
	if neverTimeout == "" {
		neverTimeout = "2s"
	}
	if consistentlyTimeout == "" {
		consistentlyTimeout = "5s"
	}
	if requireKTM == "" {
		requireKTM = "false"
	}
	requireKTMParsed, requireKTMParseErr := strconv.ParseBool(requireKTM)
	if requireKTMParseErr != nil {
		panic(requireKTMParseErr)
	}
	RequireKTM = requireKTMParsed
	d, err := time.ParseDuration(eventuallyTimeout)
	if err != nil {
		panic(err)
	}
	EventuallyTimeout = d
	d, err = time.ParseDuration(neverTimeout)
	if err != nil {
		panic(err)
	}
	NeverTimeout = d
	d, err = time.ParseDuration(consistentlyTimeout)
	if err != nil {
		panic(err)
	}
	ConsistentlyTimeout = d
}
