// Licensed to Apache Software Foundation (ASF) under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Apache Software Foundation (ASF) licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

syntax = "proto3";

package banyandb.stream.v1;

import "banyandb/common/v1/common.proto";
import "banyandb/model/v1/common.proto";
import "google/protobuf/timestamp.proto";
import "validate/validate.proto";

option go_package = "github.com/apache/skywalking-banyandb/api/proto/banyandb/stream/v1";
option java_package = "org.apache.skywalking.banyandb.stream.v1";

// TagFamilySpec defines the specification of a tag family.
message TagFamilySpec {
  // name of the tag family
  string name = 1;
  // names of tags in the tag family
  repeated string tag_names = 2;
}

message ElementValue {
  // element_id could be span_id of a Span or segment_id of a Segment in the context of stream
  string element_id = 1;
  // timestamp is in the timeunit of milliseconds. It represents
  // 1) either the start time of a Span/Segment,
  // 2) or the timestamp of a log
  google.protobuf.Timestamp timestamp = 2;
  // the order of tag_families' items match TagFamilySpec
  repeated model.v1.TagFamilyForWrite tag_families = 3;
}

message WriteRequest {
  // the metadata is required only for the first request of gRPC stream.
  common.v1.Metadata metadata = 1;
  // the element is required.
  ElementValue element = 2 [(validate.rules).message.required = true];
  // the message_id is required.
  uint64 message_id = 3 [(validate.rules).uint64.gt = 0];
  // the tag family specification.
  // If this is not set with the indicated metadata, use the schema definition.
  // If this is not set, use the existing spec declaration from previous requests in the current gRPC stream.
  repeated TagFamilySpec tag_family_spec = 4;
}

message WriteResponse {
  // the message_id from request.
  uint64 message_id = 1 [(validate.rules).uint64.gt = 0];
  // status indicates the request processing result
  string status = 2;
  // the metadata from request when request fails
  common.v1.Metadata metadata = 3;
}

message InternalWriteRequest {
  uint32 shard_id = 1;
  repeated model.v1.TagValue entity_values = 2;
  WriteRequest request = 3;
}
