// Licensed to Apache Software Foundation (ASF) under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Apache Software Foundation (ASF) licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

syntax = "proto3";

package banyandb.property.v1;

option go_package = "github.com/apache/skywalking-banyandb/api/proto/banyandb/property/v1";
option java_package = "org.apache.skywalking.banyandb.property.v1";

message PropagationContext {
  repeated string nodes = 1;
  int32 max_propagation_count = 2;
  int32 current_propagation_count = 3;
  string origin_node = 4;
}

message PropagationTraceContext {
  string trace_id = 1;
  string parent_span_id = 2;
}

message PropagationRequest {
  PropagationContext context = 1;
  PropagationTraceContext trace_context = 2;
  string group = 3;
  uint32 shard_id = 4;
}

message PropagationResponse {}

service GossipService {
  // Propagation is used to propagate messages across nodes in the cluster.
  rpc Propagation(PropagationRequest) returns (PropagationResponse) {}
}
