// Licensed to Apache Software Foundation (ASF) under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Apache Software Foundation (ASF) licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

syntax = "proto3";

package banyandb.measure.v1;

import "banyandb/measure/v1/query.proto";
import "banyandb/measure/v1/topn.proto";
import "banyandb/measure/v1/write.proto";
import "google/api/annotations.proto";
import "protoc-gen-openapiv2/options/annotations.proto";

option go_package = "github.com/apache/skywalking-banyandb/api/proto/banyandb/measure/v1";
option java_package = "org.apache.skywalking.banyandb.measure.v1";
option (grpc.gateway.protoc_gen_openapiv2.options.openapiv2_swagger) = {base_path: "/api"};

message DeleteExpiredSegmentsRequest {
  string group = 1;
  repeated string segment_suffixes = 2;
}

message DeleteExpiredSegmentsResponse {
  int64 deleted = 1;
}

service MeasureService {
  rpc Query(QueryRequest) returns (QueryResponse) {
    option (google.api.http) = {
      post: "/v1/measure/data"
      body: "*"
    };
  }

  // InternalQuery is used for internal distributed query between liaison and data nodes.
  // Returns InternalQueryResponse with shard information for proper deduplication.
  rpc InternalQuery(InternalQueryRequest) returns (InternalQueryResponse);

  rpc Write(stream WriteRequest) returns (stream WriteResponse);
  rpc TopN(TopNRequest) returns (TopNResponse) {
    option (google.api.http) = {
      post: "/v1/measure/topn"
      body: "*"
    };
  }
  rpc DeleteExpiredSegments(DeleteExpiredSegmentsRequest) returns (DeleteExpiredSegmentsResponse);
}
