// Licensed to Apache Software Foundation (ASF) under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Apache Software Foundation (ASF) licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

package observability

import (
	"github.com/apache/skywalking-banyandb/pkg/meter"
	"github.com/apache/skywalking-banyandb/pkg/run"
)

// BypassRegistry is a MetricsRegistry that does not collect any metrics.
var BypassRegistry = NewBypassRegistry()

type bypassRegistry struct{}

// NewBypassRegistry creates a new instance of bypassMetricsRegistry.
func NewBypassRegistry() MetricsRegistry {
	return &bypassRegistry{}
}

func (b *bypassRegistry) Name() string {
	return "bypass-metrics-registry"
}

func (b *bypassRegistry) Serve() run.StopNotify {
	return nil
}

func (b *bypassRegistry) GracefulStop() {
}

func (b *bypassRegistry) With(_ meter.Scope) Factory {
	return &bypassFactory{}
}

func (b *bypassRegistry) NativeEnabled() bool {
	return false
}

// bypassFactory is a Factory implementation that does nothing.
type bypassFactory struct{}

func (b *bypassFactory) NewCounter(_ string, _ ...string) meter.Counter {
	return &bypassCounter{}
}

func (b *bypassFactory) NewGauge(_ string, _ ...string) meter.Gauge {
	return &bypassGauge{}
}

func (b *bypassFactory) NewHistogram(_ string, _ meter.Buckets, _ ...string) meter.Histogram {
	return &bypassHistogram{}
}

// bypassCounter is a Counter that does nothing.
type bypassCounter struct{}

func (b *bypassCounter) Inc(_ float64, _ ...string) {}

func (b *bypassCounter) Delete(_ ...string) bool {
	return true
}

// bypassGauge is a Gauge that does nothing.
type bypassGauge struct{}

func (b *bypassGauge) Set(_ float64, _ ...string) {}

func (b *bypassGauge) Add(_ float64, _ ...string) {}

func (b *bypassGauge) Delete(_ ...string) bool {
	return true
}

// bypassHistogram is a Histogram that does nothing.
type bypassHistogram struct{}

func (b *bypassHistogram) Observe(_ float64, _ ...string) {}

func (b *bypassHistogram) Delete(_ ...string) bool {
	return true
}
