// Licensed to Apache Software Foundation (ASF) under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Apache Software Foundation (ASF) licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

syntax = "proto3";

package banyandb.property.v1;

import "banyandb/common/v1/common.proto";
import "banyandb/model/v1/query.proto";
import "google/protobuf/timestamp.proto";
import "validate/validate.proto";

option go_package = "github.com/apache/skywalking-banyandb/api/proto/banyandb/property/v1";
option java_package = "org.apache.skywalking.banyandb.property.v1";

// Property stores the user defined data
message Property {
  // metadata is the identity of a property
  common.v1.Metadata metadata = 1 [(validate.rules).message.required = true];
  // id is the identity of a property
  string id = 2 [(validate.rules).string.min_len = 1];
  // tag stores the content of a property
  repeated model.v1.Tag tags = 3 [(validate.rules).repeated.min_items = 1];
  // updated_at indicates when the property is updated
  google.protobuf.Timestamp updated_at = 4;
}
