// Licensed to Apache Software Foundation (ASF) under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Apache Software Foundation (ASF) licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

// Package route defines interfaces and types for providing route table information.
package route

import (
	databasev1 "github.com/apache/skywalking-banyandb/api/proto/banyandb/database/v1"
)

// TableProvider provides route table information for the cluster state API.
// Implementations should return a RouteTable with all registered nodes and their health states.
type TableProvider interface {
	// GetRouteTable returns the current route table state.
	// The returned RouteTable contains:
	// - Registered: all nodes known to this provider (full Node information)
	// - Active: node names (Node.Metadata.Name) that are currently healthy
	// - Evictable: node names (Node.Metadata.Name) that are unhealthy and being retried
	// The returned RouteTable is a copy and safe for concurrent access.
	GetRouteTable() *databasev1.RouteTable
}
