// Licensed to Apache Software Foundation (ASF) under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Apache Software Foundation (ASF) licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

syntax = "proto3";

package banyandb.model.v1;

option go_package = "github.com/apache/skywalking-banyandb/api/proto/banyandb/model/v1";
option java_package = "org.apache.skywalking.banyandb.model.v1";

// Status is the response status for write
enum Status {
  STATUS_UNSPECIFIED = 0;
  STATUS_SUCCEED = 1;
  STATUS_INVALID_TIMESTAMP = 2;
  STATUS_NOT_FOUND = 3;
  STATUS_EXPIRED_SCHEMA = 4;
  STATUS_INTERNAL_ERROR = 5;
  STATUS_DISK_FULL = 6;
  STATUS_VERSION_UNSUPPORTED = 7; // Client version not supported
  STATUS_VERSION_DEPRECATED = 8; // Client version deprecated but still supported
  STATUS_METADATA_REQUIRED = 9; // Metadata is required for the first request
}
