// Licensed to Apache Software Foundation (ASF) under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Apache Software Foundation (ASF) licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

//go:build linux

package ebpf

import (
	"fmt"
	"os"
	"path/filepath"
)

// resolveCgroupPath validates that the given path exists and looks like a
// cgroup v2 directory (contains cgroup.procs).
func resolveCgroupPath(path string) (string, error) {
	abs, absErr := filepath.Abs(path)
	if absErr != nil {
		return "", absErr
	}
	if _, statErr := os.Stat(filepath.Join(abs, "cgroup.procs")); statErr != nil {
		return "", fmt.Errorf("not a valid cgroup directory: %w", statErr)
	}
	return abs, nil
}
